/* Mehrspieler-Rasterbike fr via Netzwerk: Server
    2000 by Andreas Meile, CH-8242 Hofen SH */

#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

/* Spielstandbeschreibung */

#define RCV_BUF          20  /* Empfangspuffer */
#define STD_PORT_NR   11971  /* TCP/IP-Portnummer, wo der Server horcht */
#define MAX_SPIELER      10  /* maximale Anzahl Spieler */
#define FELD_GROESSE     50  /* Grsse des Spielfeldes */
#define TEL_PUF          20  /* Grsse Telegramm-Empfangsliste */
#define SPIEL_GESCHW 500000  /* Spieltempo: 0.5 Sek. pro Schritt */

/* Differenzen zwischen verschiedenen UNIX ausgleichen */
#ifndef socklen_t
#define socklen_t int
#endif

enum zustand { TOT, LEBEND, BEREIT, INIT };

struct telegramm_puffer {
  char txt[RCV_BUF];   /* Telegramminhalt */
  int lg;              /* Lnge */
  int sp_nr;           /* Spielernummer */
};

struct spieler {
  int x, y,   /* Position des Spielers im Feld */
      r,      /* Spielerrichtung */
      fd,     /* Filedeskriptor des zugehrigen TCP/IP-Sockets */
      anz_by; /* Anzahl Bytes im Socketpuffer */
  enum zustand zust; /* Flag, ob Spieler lebend; zu Beginn auch fr
                     Bereit-Zustandsabfrage, ob Raster"/"Bike"-Handshake
                     schon ordnungsgemss erfolgt ist */
  unsigned char sock_puf[RCV_BUF];  /* Empfangspuffer */
  struct timeval anmeld_tout;  /* Timeout bei Anmeldung */
};

const int sx[] = { 1,  0, -1, 0 };
const int sy[] = { 0, -1,  0, 1 };

struct telegramm_puffer tpuf[TEL_PUF];

struct spieler sp[MAX_SPIELER];

int spielfeld[FELD_GROESSE + 2][FELD_GROESSE + 2],  /* Spielfeldzustand */
    n_spieler,           /* Anzahl teilnehmende Spieler */
    n_unguelt,           /* davon ungltig, deren TCP/IP-Sockets geschlossen
                            wurden (sd = -1) */
    sd_acpt,             /* Accept-Socket */
    fd_max,              /* fr select() bentigt */
    rang[MAX_SPIELER],   /* Spieler-Rnge */
    naechst_rang,        /* nchster Rang */
    n_tel,               /* nchstes zu bearbeitendes Telegramm */
    master_sp,           /* Masterspieler */
    nf_tel;              /* nchster freier Platz */

struct timeval n_ereig;  /* Stattfindung nchstes Ereignis */
struct timezone zz;      /* nicht verwertet */

char *prg_name;

struct sockaddr_in insock;
socklen_t insock_gr;

/* Allgemeine Fehlerbehandlung */
void fehler(char *grund) {
  fprintf(stderr, "Fehler bei %s\n", grund);
  perror(prg_name);
  exit(1);
}

/* Zeitarithmetik */

void inkremiere_zeit(struct timeval *zw, long diff) {
  zw->tv_usec += diff;
  zw->tv_sec += zw->tv_usec / 1000000;
  zw->tv_usec %= 1000000;
}

void zeit_diff(const struct timeval *abs1,
               const struct timeval *abs2, struct timeval *erg) {
  erg->tv_sec = abs1->tv_sec - abs2->tv_sec;
  erg->tv_usec = abs1->tv_usec - abs2->tv_usec;
  if(erg->tv_usec < 0) {
    erg->tv_sec--;
    erg->tv_usec+=1000000;
  }
}

void initialisiere_spielfeld(void) {
  int i, j;
  for(i = 0; i < FELD_GROESSE + 2; i++) {
    spielfeld[i][0] = 1;
    spielfeld[i][FELD_GROESSE + 1] = 1;
  }
  for(i = 1; i <= FELD_GROESSE; i++) {
    spielfeld[0][i] = 1;
    for(j = 1; j <= FELD_GROESSE; j++)
      spielfeld[j][i] = 0;
    spielfeld[FELD_GROESSE + 1][i] = 1;
  }
}

/* neuer Spieler einbinden */
void neuer_spieler(int sd_neu) {
  int abstand_nok, i, dx, dy;

  do {
    /* Startposition zufllig! */
    sp[n_spieler].x = 5 + rand() % (FELD_GROESSE - 8);
    sp[n_spieler].y = 5 + rand() % (FELD_GROESSE - 8);
    /* Abstand sollte gengend gross sein zu den anderen! */
    abstand_nok = 0;
    for(i=0; i<n_spieler; i++) {
      if(sp[i].fd != -1) {
        dx = sp[i].x - sp[n_spieler].x;
        dy = sp[i].y - sp[n_spieler].y;
        if(dx * dx + dy * dy < 150)
          abstand_nok = 1;
      }
    }
  } while(abstand_nok);
  sp[n_spieler].r = rand() & 3; /* Richtung */
  sp[n_spieler].fd = sd_neu;
  sp[n_spieler].anz_by = 0;
  sp[n_spieler].zust = INIT;  /* INIT=frisch per accept() hineingekommen */
  if(gettimeofday(&sp[n_spieler].anmeld_tout, &zz) != 0)
    fehler("Zeit ermitteln");
  inkremiere_zeit(&sp[n_spieler].anmeld_tout, 30000000);
  /* INIT muss innerhalb von 30 s stattfinden! */
  if(sd_neu >= fd_max)
    fd_max = sd_neu + 1;
  n_spieler++;
  naechst_rang++;
  printf("Verbindung aufgenommen von Spieler %d\n", n_spieler);
}

/* Serverseitiges Socket-Schliessen: Empfangspuffer leerlesen */
void schliesse(int sd) {
  char dummy[20];
  int fcntl_flags, a_by;
  extern int errno;

  if(sd == -1)
    fprintf(stderr, "Warnung: Zweites Schliessen ignoriert\n");
  else {
    /* Empfangene Bytes leerlesen */
    fcntl_flags = fcntl(sd, F_GETFL);
    if(fcntl_flags == -1)
      fehler("Flags lesen");
    fcntl_flags |= O_NONBLOCK;
    if(fcntl(sd, F_SETFL, fcntl_flags) != 0)
      fehler("Flags setzen");
    do {
      a_by = read(sd, dummy, sizeof(dummy));
    } while(a_by > 0);
    if(a_by == -1 && errno != EWOULDBLOCK)
      fehler("Leerlesen");
    /* Puffer jetzt leer, es kann geschlossen werden */
    if(close(sd) != 0)
      fehler("Schliessen (Abweisen)");
  }
}

void schreibe_telegramm(int sd, char *telegramm, int telegr_l) {
  int i, j;
  unsigned char buf[48];

  if(sd == -1)
    fprintf(stderr, "Warnung: berflssiges Telegramm ignoriert\n");
  else {
    for(i=0; i<telegr_l; i++) {
      buf[i] = telegramm[i];
    }
    buf[telegr_l - 1] |= 0x80;  /* Kennung fr Telegrammende */
    i=0;
    while(i < telegr_l) {
      j = write(sd, buf + i, sizeof(char) * (telegr_l - i));
      if(j < 0)
        fehler("berschreiben des Sockets");
      i += j;    
    }
  }
}

void schreibe_telegramm_an_alle(char *telegramm, int telegr_l, int ausgen) {
  int i;

  for(i=0; i<n_spieler; i++)
    if(sp[i].fd != -1 && sp[i].zust != INIT && i != ausgen)
      schreibe_telegramm(sp[i].fd, telegramm, telegr_l);
}

/* Meldung, dass fertig */
void melde_fertig(int sp, char typ, int rang, int ausg) {
  char tel[3];
  tel[0] = typ;
  tel[1] = rang;
  tel[2] = 4 * sp;
  schreibe_telegramm_an_alle(tel, 3, ausg);
}

/* Position allen oder nur einem Spieler mitteilen */

void melde_position(int s) {
  /* s=-1 => alle Positionen an alle Spieler, whrend dem Spiel
             der Fall, sonst nur derjenige Spieler, der neu dazugekommen
             ist */
  char mldpuff[3 * MAX_SPIELER + 1], *t;
  int i, l;

  mldpuff[0]='P';
  l=1;
  for(i=0; i<n_spieler; i++)
    if(sp[i].fd != -1 && sp[i].zust != INIT && sp[i].zust != TOT) {
      if(i == s)
        t = mldpuff + l;
      mldpuff[l] = sp[i].x;
      mldpuff[l+1] = sp[i].y;
      mldpuff[l+2] = i << 2 | sp[i].r;
      l += 3;
    }
  if(s == -1)
    schreibe_telegramm_an_alle(mldpuff, l, -1);
  else {
    schreibe_telegramm(sp[s].fd, mldpuff, l);
    t--;
    *t='P';
    schreibe_telegramm_an_alle(t, 4, s);
  }
}

void melde_richtung(int s) {
  char mld;
  
  mld = s << 2 | sp[s].r;
  schreibe_telegramm_an_alle(&mld, 1, -1);
}

void aktualisiere_master_spieler() {
  int i;

  for(i=0; i<n_spieler && sp[i].fd == -1; i++);
  if(i < n_spieler && sp[i].zust == BEREIT && i != master_sp) {
    master_sp = i;
    schreibe_telegramm(sp[i].fd, "MA", 2);
  }
}

void setze_anmelde_timeout() {
  int i=0;

  for(i=0; i<n_spieler && (sp[i].fd == -1 || sp[i].zust != INIT); i++);
  if(i == n_spieler)
    n_ereig.tv_usec = -1;
  else
    n_ereig = sp[i].anmeld_tout;
}

/* Socket von Spieler lesen und dabei erhaltenes Telegramm zurckliefern */
int lies_telegramm(int modus) {
  fd_set warte_liste;    /* fr gleichzeitiges Warten */
  int h, h2, i, j, temp_sd, n_sel;
  size_t a_by;
  struct timeval akt_z, wart_z;

  n_sel = 1;
  while(nf_tel == n_tel && n_sel != 0) {
    FD_ZERO(&warte_liste);
    if(n_ereig.tv_usec != -1) {
      if(gettimeofday(&akt_z, &zz) != 0)
        fehler("Zeit ermitteln");
      zeit_diff(&n_ereig, &akt_z, &wart_z);
    }
    if(n_ereig.tv_usec == -1 || wart_z.tv_sec > 0 || wart_z.tv_sec == 0 &&
    wart_z.tv_usec > 0) {
      FD_SET(sd_acpt, &warte_liste);
      for(i=0; i<n_spieler; i++)
        if(sp[i].fd != -1)
          FD_SET(sp[i].fd, &warte_liste);
      n_sel = select(fd_max, &warte_liste, NULL, NULL, n_ereig.tv_usec != -1 ?
      &wart_z : NULL);
    } else
      n_sel = 0;  /* simuliere select() mit Time Out */
    if(n_sel < 0)
      fehler("Abwarten");
    if(FD_ISSET(sd_acpt, &warte_liste)) {
      insock_gr = sizeof(insock);
      temp_sd = accept(sd_acpt, (struct sockaddr *)&insock, &insock_gr);
      if(temp_sd < 1)
        fehler("Verbindungsannahme");
      if(modus == 1 && n_spieler < MAX_SPIELER) {
        neuer_spieler(temp_sd);  /* Spieler aufnehmen */
        setze_anmelde_timeout();
      } else {
        /* Einladungsphase beendet oder alles schon daran! */
        schreibe_telegramm(temp_sd, "busy", 4);
        schliesse(temp_sd); /* der Nachzgler-Client wird "abgewimmelt" */
        printf("Potentieller Spieler abgewiesen\n");
      }
    }
    /* Einzelne Spieler durchgehen */
    for(i=0; i<n_spieler; i++) {
      h = sp[i].fd;
      if(h != -1 && FD_ISSET(h, &warte_liste)) {
        a_by = read(h, sp[i].sock_puf + sp[i].anz_by, sizeof(sp[0].sock_puf)
               - sp[i].anz_by);
        if(a_by < 0)
          fehler("Telegramm lesen");
        if(a_by == 0) {
          /* Client hat sich getrennt */
          if(modus != 2)
            fprintf(stderr, "Spieler %d hat sich unerwartet getrennt\n", i+1);
          else
            printf("Spieler %d clientseitig getrennt\n", i+1);
          if(close(sp[i].fd) != 0)
            fehler("Schliessen (Client)");
          sp[i].fd = -1;
          n_unguelt++;
          if(modus == 1)
            aktualisiere_master_spieler();  /* falls bisheriger Master
            "gestorben" ist, bekommt der unmittelbar nchste Spieler diese
            Funktion */
          if(modus == 2)
            n_sel = 0; /* Simuliere Time Out */
          else {
            melde_fertig(i, 'X', naechst_rang, -1);
            naechst_rang--;
            rang[naechst_rang] = i;
          }
        }
        /* Scannen nach Telegrammende */
        h = sp[i].anz_by + a_by;
        h2 = 0;
        for(j = sp[i].anz_by; j < h; j++)
          if(sp[i].sock_puf[j] & 0x80) {
            /* Vollstndiges Telegramm gefunden => in Liste stellen */
            tpuf[nf_tel].lg = j + 1 - h2;
            memcpy(tpuf[nf_tel].txt, sp[i].sock_puf + h2, tpuf[nf_tel].lg);
            tpuf[nf_tel].txt[j - h2] &= 0x7f;
            tpuf[nf_tel].sp_nr = i;
            h2 = j + 1;
            nf_tel = (nf_tel + 1) % TEL_PUF;
            if(nf_tel == n_tel)
              fprintf(stderr,
              "WARNUNG: Telegramm-Pufferberlauf ist aufgetreten\n");
          }
        sp[i].anz_by = h - h2;
        memmove(sp[i].sock_puf, sp[i].sock_puf + h2, h - h2);
      }
      if(sp[i].anz_by == RCV_BUF) {
        fprintf(stderr, "Warnung: berlanges Telegramm bei %d erhalten =>"
        " verworfen\n", i+1);
        sp[i].anz_by = 0;
      }
    }
  }
  if(n_sel != 0) {
    i = n_tel;
    n_tel = (n_tel + 1) % TEL_PUF;
  } else
    i = -1;   /* kein Telegramm erhalten, dafr Aktion wie Mofas um 1
                 Position verschieben ausfhren */
  return i;
}

int main(int argc, char *argv[]) {
  int n_bereit,   /* fr Bereitschaftsverwaltung verwendet */
      anz_rang,   /* Grsse der Rangliste */
      tel,        /* Telegramm */
      drinbleib,  /* fr Schleife */
      s,          /* Spieler */
      port_nr,    /* effektive TCP/IP-Portnummer */
      i;          /* Hilfsvariable */

  struct timeval zufallsg_init;

  char bike_tel[30]; /* Puffer fr Telegramm */

  printf("Rasterbike-Server V1.01  2000 by Andreas Meile, CH-8242 Hofen SH\n");
  prg_name = argv[0];

  switch(argc) {
    case 1:
      port_nr = STD_PORT_NR;
      break;
    case 2:
      if(sscanf(argv[1], "%d", &port_nr) == 1)
        break;
    default:
      fprintf(stderr, "Aufrufformat: %s [ <Port-Nummer> ]\n", argv[0]);
      exit(1);
  }

  initialisiere_spielfeld();
  n_tel = 0;   /* Zu Beginn leere Telegrammliste */
  nf_tel = 0;

  /* Zufallsgenerator initialisieren */
  if(gettimeofday(&zufallsg_init, &zz) != 0)
    fehler("Zeit ermitteln");
  srand(zufallsg_init.tv_sec ^ zufallsg_init.tv_usec);


  /* Socket erzeugen */
  sd_acpt = socket(AF_INET, SOCK_STREAM, 0);
  if(sd_acpt == -1)
    fehler("Socket erzeugen");
  fd_max = sd_acpt + 1;

  /* An Portnummer anbinden */
  memset(&insock, 0, sizeof(insock));
  insock.sin_family = AF_INET;
  insock.sin_port   = htons(port_nr);
  if (bind(sd_acpt, (struct sockaddr *)&insock, sizeof(insock)) == -1)
    fehler("Anbinden");

  /* Warten und "lose" */
  if (listen(sd_acpt, 5) == -1)
    fehler("Horchen");

  n_ereig.tv_usec = -1;  /* keine Uhr */  
  /* Phase 1: auf Spieler warten entgegennehmen */
  n_spieler = 0;
  n_bereit = 0;
  n_unguelt = 0;
  drinbleib = 1;
  master_sp = -1;
  printf("Server auf Port %d gestartet - Einladungsphase\n", port_nr);
  while(drinbleib) {
    tel = lies_telegramm(1);
    if(tel == -1) {
      /* Anmeldefrist ist abgelaufen! */
      for(s=0; s<n_spieler && (sp[s].fd == -1 || sp[s].zust != INIT); s++);
      assert(s < n_spieler); /* Es muss ein anmeldender Spieler existieren! */
      schliesse(sp[s].fd);
      sp[s].fd = -1;
      n_unguelt++;
      printf("Anmeldefrist fr Spieler %d abgelaufen => zurckgewiesen\n",
      s+1);
      setze_anmelde_timeout();
      aktualisiere_master_spieler();
    } else {
      s = tpuf[tel].sp_nr;
      switch(sp[s].zust) {
        case INIT:
          if(tpuf[tel].lg == 6 && !strncmp(tpuf[tel].txt, "Raster", 6)) {
            /* Handshake: Kennung, um irrtmliches "telnet xy.z.ch 11971"
               oder http://xy.z.ch:11971 unterscheiden zu knnen im Sinne einer
               Kennung wie z.B. "BM" im BMP-Dateiformat */
            n_bereit++;
            sp[s].zust = BEREIT;
            strcpy(bike_tel, "Bike");
            bike_tel[4] = s;  /* Spielernummer dem Client mitteilen */
            schreibe_telegramm(sp[s].fd, bike_tel, 5);
            /* Meldung an brige, dass neuer Spieler */
            melde_position(s);
            spielfeld[sp[s].x][sp[s].y] = 1;
            setze_anmelde_timeout();
            printf("Spieler %d verbunden\n", s+1);
          } else {
            fprintf(stderr, "Protokollfehler: Kein Rasterbike-Client!\n"
                            "Spieler %d abgewiesen\n", s+1);
            schreibe_telegramm(sp[s].fd, "INVALID CLIENT ", 15);
            schliesse(sp[s].fd);
            n_unguelt++;
            sp[s].fd = -1;
          }
          aktualisiere_master_spieler();
        break;
        case BEREIT:
          if(tpuf[tel].lg == 2 && !strncmp(tpuf[tel].txt, "EI", 2)) {
            if(s == master_sp) {
              /* Einladungsphase beenden: Nur der erste Spieler
                 kann dies tun, da er Masterspieler ist! */
              if(n_bereit + n_unguelt == n_spieler)
                drinbleib = 0;
              else {
                /* Halt! Noch nicht alle Spieler sind initialisiert =>
                   "Not Yet" als NAK dem Master senden */
                printf("Einladungsphase noch nicht beendbar\n");
                schreibe_telegramm(sp[master_sp].fd, "NY", 2);
              }
            } else
              fprintf(stderr, "Fehler bei EI: Spieler %d ist nicht Master\n",
              s+1);
          } else {
            fprintf(stderr, "Protokollfehler bei %d: EI vom Master erwartet\n",
            s+1);
          }
        break;
        default:
          fprintf(stderr, "Interner Zustandsfehler bei %s\n", s+1);
          exit(1);
      }
    }
  }
  /* Jedem ein "Please Prepare" senden, auch an den Master als ACK */
  printf("Einladungsphase beendet -> Bereitmachungsphase\n");
  schreibe_telegramm_an_alle("PP", 2, -1);
  /* Phase 2: Von jedem Spieler wird ein Tastendruck erwartet, dass er
     bereit ist */
  n_bereit = 0;
  while(n_bereit + n_unguelt < n_spieler) {
    tel = lies_telegramm(0);
    s = tpuf[tel].sp_nr;
    switch(sp[s].zust) {
      case BEREIT:
        if(tpuf[tel].lg == 1 && tpuf[tel].txt[0] == 'R') {
          sp[s].zust = LEBEND;
          n_bereit++;
          melde_richtung(s);
          printf("Spieler %d hat Bereitschaft besttigt\n", s+1);
        } else {
          fprintf(stderr, "Protokollfehler bei %d: Bereitschaft R erwartet\n",
          s+1);
        }
      break;
      case LEBEND:
        fprintf(stderr, "Protokollfehler bei %d: Bereits bereit\n", s+1);
      break;
      default:
        fprintf(stderr, "Interner Zustandsfehler bei %d\n", s+1);
        exit(1);
    }
  }
  printf("Alle Spieler sind bereit\n");
  /* Zeit initialisieren */
  if(gettimeofday(&n_ereig, &zz) != 0)
    fehler("Zeit auslesen");
  /* Count-Down "Achtung! Fertig! Los!" durchgeben */
  for(i=0; i<3; i++) {
    inkremiere_zeit(&n_ereig, 800000);  /* Abstand = 0,8 s */
    while(lies_telegramm(0) > 0);   /* Ignoriere smtliche Telegramme */
    sprintf(bike_tel, "A%d", i);
    schreibe_telegramm_an_alle(bike_tel, 2, -1);
  }
  printf("Spielrunde gestartet\n");
  /* Ab hier eigentliches Spiel */
  anz_rang = naechst_rang = n_spieler - n_unguelt;
  inkremiere_zeit(&n_ereig, SPIEL_GESCHW);
  while(naechst_rang > 0) {  /* Solange noch jemand herumkurvt */
    i = lies_telegramm(0);
    if(i != -1) {
      /* Lenkkommando eines Spielers */
      s = tpuf[i].sp_nr;
      if(sp[s].zust == LEBEND) {
        if(tpuf[i].lg == 1) {
          if(tpuf[i].txt[0] >= '@' && tpuf[i].txt[0] <= 'C') {
            sp[s].r = tpuf[i].txt[0] & 0x3;
            melde_richtung(s);
          } else
            fprintf(stderr, "Nicht verstandenes Telegramm von %d : %c\n", s+1,
            tpuf[i].txt[0]);
        } else
          fprintf(stderr, "Falsches Telegramm von %d (L=%d)\n", s+1,
          tpuf[i].lg);
      }
    } else {
      /* Alle Mofas um 1 Position vorwrtsbewegen */
      for(i=0; i<n_spieler; i++) {
        if(sp[i].fd != -1 && sp[i].zust == LEBEND) {
          sp[i].x += sx[sp[i].r];
          sp[i].y += sy[sp[i].r];
        }
      }
      melde_position(-1);
      /* auf Kollisionen prfen */
      for(i=0; i<n_spieler; i++) {
        if(sp[i].fd != -1 && sp[i].zust == LEBEND) {
          if(spielfeld[sp[i].x][sp[i].y]) {
            /* Png!!! */
            melde_fertig(i, 'K', naechst_rang, -1);
            sp[i].zust = TOT;
            naechst_rang--;
            rang[naechst_rang] = i;
            printf("Spieler %d ausgeschieden\n", i+1);
          } else
            spielfeld[sp[i].x][sp[i].y] = 1;
        }
      }
      inkremiere_zeit(&n_ereig, SPIEL_GESCHW);
    }
  }
  printf("Spielrunde beendet.\n");
  n_ereig.tv_usec = -1;
  /* Schlussauswertung */
  strcpy(bike_tel, "Ra");
  for(i=0; i<anz_rang; i++)
    bike_tel[2 + i] = rang[i];
  schreibe_telegramm_an_alle(bike_tel, 2 + anz_rang, -1);
  printf("Rangliste bermittelt\n");
  /* Ausklang: Jeder Client schickt ein CA (Close ACK) */
  while(n_unguelt < n_spieler) {
    i = lies_telegramm(2);
    if(i != -1 && tpuf[i].lg == 2 && !strncmp(tpuf[i].txt, "CA", 2)) {
      s = tpuf[i].sp_nr;
      schliesse(sp[s].fd);
      sp[s].fd = -1;
      n_unguelt++;
      printf("Spieler %d durch Server getrennt\n", s+1);
    }
  }
  if(close(sd_acpt) != 0)
    fehler("Schliessen (Port)");
  printf("Rasterbike-Server beendet\n");
  return 0;
}
